@@ -1,5 +1,5 @@
 # <HTTPretty - HTTP client mock for Python>
-# Copyright (C) <2011-2020> Gabriel Falcão <gabriel@nacaolivre.org>
+# Copyright (C) <2011-2021> Gabriel Falcão <gabriel@nacaolivre.org>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation
@@ -73,6 +73,59 @@ from datetime import datetime
 from datetime import timedelta
 from errno import EAGAIN
 
+class __internals__:
+    thread_timeout = 0  # https://github.com/gabrielfalcao/HTTPretty/issues/426
+    temp_files = []
+    threads = []
+
+    @classmethod
+    def cleanup_sockets(cls):
+        cls.cleanup_temp_files()
+        cls.cleanup_threads()
+
+    @classmethod
+    def cleanup_threads(cls):
+        for t in cls.threads:
+            t.join(cls.thread_timeout)
+            if t.is_alive():
+                raise socket.timeout(cls.thread_timeout)
+
+    @classmethod
+    def create_thread(cls, *args, **kwargs):
+        return threading.Thread(*args, **kwargs)
+
+    @classmethod
+    def cleanup_temp_files(cls):
+        for fd in cls.temp_files[:]:
+            try:
+                fd.close()
+            except Exception as e:
+                logger.debug('error closing file {}: {}'.format(fd, e))
+            cls.temp_files.remove(fd)
+
+    @classmethod
+    def create_temp_file(cls):
+        fd = tempfile.TemporaryFile()
+        cls.temp_files.append(fd)
+        return fd
+
+def set_default_thread_timeout(timeout):
+    """sets the default thread timeout for HTTPretty threads
+
+    :param timeout: int
+    """
+    __internals__.thread_timeout = timeout
+
+def get_default_thread_timeout():
+    """sets the default thread timeout for HTTPretty threads
+
+    :returns: int
+    """
+
+    return __internals__.thread_timeout
+
+
+SOCKET_GLOBAL_DEFAULT_TIMEOUT = socket._GLOBAL_DEFAULT_TIMEOUT
 old_socket = socket.socket
 old_socketpair = getattr(socket, 'socketpair', None)
 old_SocketType = socket.SocketType
@@ -117,11 +170,23 @@ try:
 except ImportError:
     _ssl = None
 # used to handle error caused by ndg-httpsclient
+pyopenssl_overrides_inject = []
+pyopenssl_overrides_extract = []
 try:
     from requests.packages.urllib3.contrib.pyopenssl import inject_into_urllib3, extract_from_urllib3
-    pyopenssl_override = True
+    pyopenssl_overrides_extract.append(extract_from_urllib)
+    pyopenssl_overrides_inject.append(inject_from_urllib)
 except Exception:
-    pyopenssl_override = False
+    pass
+
+
+
+try:
+    from urllib3.contrib.pyopenssl import extract_from_urllib3, inject_into_urllib3
+    pyopenssl_overrides_extract.append(extract_from_urllib)
+    pyopenssl_overrides_inject.append(inject_from_urllib)
+except Exception:
+    pass
 
 
 try:
@@ -143,6 +208,7 @@ DEFAULT_HTTPS_PORTS = frozenset([443])
 POTENTIAL_HTTPS_PORTS = set(DEFAULT_HTTPS_PORTS)
 
 
+
 def FALLBACK_FUNCTION(x):
     return x
 
@@ -347,30 +413,65 @@ class HTTPrettyRequestEmpty(object):
     headers = EmptyRequestHeaders()
 
 
+
 class FakeSockFile(object):
     """Fake socket file descriptor. Under the hood all data is written in
     a temporary file, giving it a real file descriptor number.
 
     """
     def __init__(self):
-        self.file = tempfile.TemporaryFile()
+        self.file = None
+        self._fileno = None
+        self.__closed__ = None
+        self.reset()
+
+    def reset(self):
+        if self.file:
+            try:
+                self.file.close()
+            except Exception as e:
+                logger.debug('error closing file {}: {}'.format(self.file, e))
+            self.file = None
+
+        self.file = __internals__.create_temp_file()
         self._fileno = self.file.fileno()
+        self.__closed__ = False
 
     def getvalue(self):
         if hasattr(self.file, 'getvalue'):
-            return self.file.getvalue()
+            value = self.file.getvalue()
         else:
-            return self.file.read()
+            value = self.file.read()
+        self.file.seek(0)
+        return value
 
     def close(self):
-        self.socket.close()
-        self.file.close()
+        if self.__closed__:
+            return
+        self.__closed__ = True
+        self.flush()
+
+    def flush(self):
+        try:
+            super().flush()
+        except Exception as e:
+            logger.debug('error closing file {}: {}'.format(self, e))
+
+        try:
+            self.file.flush()
+        except Exception as e:
+            logger.debug('error closing file {}: {}'.format(self.file, e))
+
+
 
     def fileno(self):
         return self._fileno
 
     def __getattr__(self, name):
-        return getattr(self.file, name)
+        try:
+            return getattr(self.file, name)
+        except AttributeError:
+            return super().__getattribute__(name)
 
     def __del__(self):
         try:
@@ -378,6 +479,11 @@ class FakeSockFile(object):
         except (ValueError, AttributeError):
             pass
 
+        # Adding the line below as a potential fix of github issue #426
+        # that seems to be a compatible the solution of #413
+        self.file.close()
+
+
 
 class FakeSSLSocket(object):
     """Shorthand for :py:class:`~httpretty.core.fakesock`
@@ -409,6 +515,8 @@ class fakesock(object):
         """drop-in replacement for :py:class:`socket.socket`
         """
         _entry = None
+        _read_buf = None
+
         debuglevel = 0
         _sent_data = []
         is_secure = False
@@ -580,17 +688,18 @@ class fakesock(object):
             self._bufsize = bufsize
 
             if self._entry:
-                t = threading.Thread(
+                t = __internals__.create_thread(
                     target=self._entry.fill_filekind, args=(self.fd,)
                 )
+
                 t.start()
-                if self.timeout == socket._GLOBAL_DEFAULT_TIMEOUT:
-                    timeout = None
+                if self.timeout == SOCKET_GLOBAL_DEFAULT_TIMEOUT:
+                    timeout = get_default_thread_timeout()
                 else:
                     timeout = self.timeout
-                t.join(timeout)
+                t.join(None)
                 if t.is_alive():
-                    raise socket.timeout
+                    raise socket.timeout(timeout)
 
             return self.fd
 
@@ -708,7 +817,9 @@ class fakesock(object):
             matcher, entries = httpretty.match_uriinfo(info)
 
             if not entries:
+                logger.debug('no entries matching {}'.format(request))
                 self._entry = None
+                self._read_buf = None
                 self.real_sendall(data, request=request)
                 return
 
@@ -727,8 +838,9 @@ class fakesock(object):
                 if self.truesock:
                     self.truesock.settimeout(new_timeout)
 
-        def send(self, *args, **kwargs):
-            return self.forward_and_trace('send', *args, **kwargs)
+        def send(self, data, *args, **kwargs):
+            self.sendall(data, *args, **kwargs)
+            return len(data)
 
         def sendto(self, *args, **kwargs):
             return self.forward_and_trace('sendto', *args, **kwargs)
@@ -742,12 +854,20 @@ class fakesock(object):
         def recvfrom(self, *args, **kwargs):
             return self.forward_and_trace('recvfrom', *args, **kwargs)
 
-        def recv(self, buffersize=None, *args, **kwargs):
-            buffersize = buffersize or self._bufsize
-            return self.forward_and_trace('recv', buffersize, *args, **kwargs)
+        def recv(self, buffersize=0, *args, **kwargs):
+            if not self._read_buf:
+                self._read_buf = io.BytesIO()
+
+            if self._entry:
+                self._entry.fill_filekind(self._read_buf)
+
+            if not self._read_buf:
+                raise UnmockedError('socket cannot recv(): {!r}'.format(self))
+
+            return self._read_buf.read(buffersize)
 
         def __getattr__(self, name):
-            if name in ('getsockopt', ) and not self.truesock:
+            if name in ('getsockopt', 'selected_alpn_protocol') and not self.truesock:
                 self.truesock = self.create_socket()
             elif httpretty.allow_net_connect and not self.truesock:
                 # can't call self.connect_truesock() here because we
@@ -1472,6 +1592,7 @@ class httpretty(HttpBaseClass):
         cls._entries.clear()
         cls.latest_requests = []
         cls.last_request = HTTPrettyRequestEmpty()
+        __internals__.cleanup_sockets()
 
     @classmethod
     def historify_request(cls, headers, body='', sock=None):
@@ -1545,7 +1666,7 @@ class httpretty(HttpBaseClass):
         """
         uri_is_string = isinstance(uri, str)
 
-        if uri_is_string and re.search(r'^\w+://[^/]+[.]\w{2,}$', uri):
+        if uri_is_string and re.search(r'^\w+://[^/]+[.]\w{2,}(:[0-9]+)?$', uri):
             uri += '/'
 
         if isinstance(responses, list) and len(responses) > 0:
@@ -1720,9 +1841,9 @@ def apply_patch_socket():
     socket.__dict__['getaddrinfo'] = fake_getaddrinfo
 
 
-    if pyopenssl_override:
-        # Take out the pyopenssl version - use the default implementation
-        extract_from_urllib3()
+    # Take out the pyopenssl version - use the default implementation
+    for extract_from_urllib3 in pyopenssl_overrides_extract:
+        extract_into_urllib3()
 
     if requests_urllib3_connection is not None:
         urllib3_wrap = partial(fake_wrap_socket, old_requests_ssl_wrap_socket)
@@ -1795,8 +1916,9 @@ def undo_patch_socket():
         requests_urllib3_connection.__dict__['ssl_wrap_socket'] = \
             old_requests_ssl_wrap_socket
 
-    if pyopenssl_override:
-        # Put the pyopenssl version back in place
+
+    # Put the pyopenssl version back in place
+    for inject_from_urllib3 in pyopenssl_overrides_inject:
         inject_into_urllib3()
 
 
